#ifdef _WIN32
typedef unsigned char uint8_t;
#else
#include <stdint.h>
#endif

#include <cv.h>
#include <stdio.h>

#include "defs.h"
#include "image.h"
#include "main.h"

#include <math.h>

#include <cv.h>
#include <highgui.h>

#include "geometry.h"

#include "correlation.h"

static pt2d_t inverse_transform(projection_t *data,
                                pt2d_t from)
{
    pt2d_t to;
    pt3d_t xx, yy, cm;

    xx = multiply(from.x, data->vx);
    yy = multiply(from.y, data->vy);

    cm = add(data->Os, add(xx, yy));

    to.x = cm.x * data->d / cm.z;
    to.y = cm.y * data->d / cm.z;

    return to;
}

static int getpixel_fast_antialiasing(image_t *img, projection_t *data,
                                      double x, double y, int offsetx,
                                      int offsety, int *r, int *g, int *b)
{
    double i0,j0,i1,j1,I_,J_;
    int pR,pB,pG;
    double c1,c2,c3,c4;
    pt2d_t from,to;

    from.x = (float)x / (float)RESOLUTION;
    from.y = y;

    /*Point normé => point dénormé*/
    to = inverse_transform(data, from);

    /* Pourcentage de couleur */
    j0 = to.x;
    i0 = to.y;
    j1 = to.x + 1.0;
    i1 = to.y + 1.0;

    J_ = (int)to.x;
    I_ = (int)to.y;

    c1 = ((I_+1) - i0) * ((J_+1) - j0);
    c2 = ((I_+1) - i0) * (j1 - (J_+1));
    c3 = (i1 - (I_+1)) * (j1 - (J_+1));
    c4 = (i1 - (I_+1)) * ((J_+1) - j0);
    pR = pB = pG = 0;
    getpixel_fast(img,(int)to.x+offsetx,(int)to.y+offsety,r,g,b);
    pR = c1 * (*r);
    pG = c1 * (*g);
    pB = c1 * (*b);
    getpixel_fast(img,(int)to.x+1+offsetx,(int)to.y+offsety,r,g,b);
    pR += c2 * (*r);
    pG += c2 * (*g);
    pB += c2 * (*b);
    getpixel_fast(img,(int)to.x+1+offsetx,(int)to.y+1+offsety,r,g,b);
    pR += c3*(*r);
    pG += c3*(*g);
    pB += c3*(*b);
    getpixel_fast(img,(int)to.x+offsetx,(int)to.y+1+offsety,r,g,b);
    pR += c4*(*r);
    pG += c4*(*g);
    pB += c4*(*b);

    *r = pR;
    *g = pG;
    *b = pB;

    return 0;
}

/*****************************************************************************
 * Calcul_bande: calcule la distance entre deux matrices de pixels
 *               de taille CORR_X × CORR_Y
 *****************************************************************************/
static int Calcul_bande(window_t *Refer, window_t *Trans,
                        int dx, int dy, int x, int y)
{
    int i, j;
    int corr = 0;
    int delta;
    window_t P1, P2;

    for(j = -CORR_Y/2 ; j < CORR_Y-CORR_Y/2; j++)
    {
        int k = j * REFER_WIDTH + x;
        int l = (j + dy) * TRANS_WIDTH + x * RESOLUTION + dx;

        for(i = -CORR_X/2 ; i < CORR_X-CORR_X/2; i++)
        {
            if(i+x < 0 || i+x >= REFER_WIDTH-1)
            {
                P1.r = 255;
                P1.g = 0;
                P1.b = 255;
            }
            else
            {
                P1 = *(Refer+k+i);
            }

            if(x+i+dx < 0 || i+x*RESOLUTION+dx >= TRANS_WIDTH-1)
            {
                P2.r = 255;
                P2.g = 0;
                P2.b = 255;
            }
            else
            {
                P2 = *(Trans+l+i);
            }

            /* Distance euclidienne */
            delta = sqrt(sq(P1.r-P2.r) + sq(P1.g-P2.g) + sq(P1.b-P2.b));
            corr += delta;
        }
    }

    return corr;
}

/*****************************************************************************
 * Translation_bande: calcule la meilleure corrélation d'une matrice de
 *                    taille CORR_X × CORR_Y dans une fenêtre de taille
 *                    TRANS_X × TRANS_Y.
 *****************************************************************************/
int Translation_bande(int x, int y, window_t *Refer,
                      window_t *Trans, pixeldepth_t *Map_norm)
{
    int dx, dy;
    int corr, corr_min = 0x10000000;
    int dx_min = 0;
    int dold_x = 0, dold_y = 0;
    int good_x = 0, good_y = 0;

    /* Premiere colonne */
    if(x > 0)
    {
	/*Récupération du pixel précédent de la même ligne*/
        dold_x = (Map_norm+1*REFER_WIDTH+x-1)->depth;
        if((Map_norm+1*REFER_WIDTH+x-1)->corr < THRESH)
            good_x = 1;
    }

    /* Premiere ligne */
    if(y > 2)
    {
        /* Récupération du pixel précédent de la même colonne */
        dold_y = (Map_norm+x)->depth;
        if((Map_norm+x)->corr < THRESH)
            good_y = 1;
    }

    /* Parcours de la translation de gauche a droite */
    for(dy = -TRANS_Y/2; dy < TRANS_Y-TRANS_Y/2; dy++)
    {
        for(dx = -TRANS_X/2; dx < TRANS_X-TRANS_X/2; dx++)
        {
            corr = 16 * Calcul_bande(Refer, Trans, dx, dy, x, y);
            if(good_x)
                corr += X_WEIGHT * (abs(dx - dold_x));

            if(good_y)
                corr += Y_WEIGHT * (abs(dx - dold_y));
	    (Map_norm+REFER_WIDTH+x)->tabcorr[dx+(TRANS_X/2)]  = corr;
            /* Minimisation de la corrélation */
            if(corr < corr_min)
            {
                corr_min = corr;
                dx_min = dx;
            }
        }
    }
   (Map_norm+1*REFER_WIDTH+x)->depth = dx_min;
   (Map_norm+1*REFER_WIDTH+x)->corr = corr_min;
   return corr_min;
}

int Correlation_bande(image_t *left , projection_t *left_norm_data,
                      image_t *right, projection_t *right_norm_data,
                      image_t *left_depthmap_norm,
                      image_t *right_depthmap_norm, double adjust,map_t *Map)
{
    int x, y, i;
    int corr, progress;
    window_t *lines_refer, *lines_trans;
    map_t *Map_norm;
    pt2d_t pt2d;
    
    
    /*Détermination de la taille de la bande de référence de 3 lignes*/
//    trans_width = (right_norm_data->right-right_norm_data->left)*RESOLUTION + 1;
//    refer_width = left_norm_data->right - left_norm_data->left + 1;
    

    /*Allocation des lignes normalisée de 3 lignes*/
    lines_refer = (window_t*)malloc(sizeof(window_t)*(REFER_WIDTH*CORR_Y+1));
    lines_trans = (window_t*)malloc(sizeof(window_t)*(TRANS_WIDTH*WINTRANS_Y+1));

    /*Allocation de la 'Map' normé de trois lignes*/
    Map_norm = malloc(sizeof(map_t));
    Map_norm->elem = malloc(sizeof(pixeldepth_t) * (2 * REFER_WIDTH));
   for(y=0 ; y<2 ; y++)
    {
	for(x=0 ; x<REFER_WIDTH ; x++)
	{
	    memset((Map_norm->elem+x+REFER_WIDTH*y)->tabcorr, 0, TRANS_X * sizeof(int));
	}
    }
 

    /*Initialisation des images resultat en bleu*/
    cvFillImage(left_depthmap_norm->_private, CV_RGB(0,0,255));
    cvFillImage(right_depthmap_norm->_private, CV_RGB(0,0,255));
    /*Parcours de l'image normalisé*/
    for(progress=0,y=0 ; y<left_norm_data->bottom-left_norm_data->top; y++)
    {
        while( y * 100 / (left_norm_data->bottom-left_norm_data->top) >= progress)
        {
            printf("\r %i %% effectués", progress);
            fflush(stdout);
            progress ++;
        }
        /* Décalage de ligne */
        for(i = 0; i < CORR_Y - 1; i++ )
            memcpy(lines_refer + REFER_WIDTH * i,
                   lines_refer + REFER_WIDTH * (i + 1),
                   sizeof(window_t) * REFER_WIDTH);

        for(i = 0; i < WINTRANS_Y - 1; i++ )
            memcpy(lines_trans + TRANS_WIDTH * i,
                   lines_trans + TRANS_WIDTH * (i + 1),
                   sizeof(window_t) * TRANS_WIDTH);
        /* Lecture de la ligne suivante de l'image gauche*/
        for(x = 0 ; x < REFER_WIDTH; x++)
        {
	    getpixel_fast_antialiasing(left, left_norm_data, 
	                               (x+(int)left_norm_data->left)*RESOLUTION ,
                                        y+(int)left_norm_data->top+adjust, 
				       0, 0,
                                       &(lines_refer+x+(CORR_Y-1)*REFER_WIDTH)->r,
                                       &(lines_refer+x+(CORR_Y-1)*REFER_WIDTH)->g,
                                       &(lines_refer+x+(CORR_Y-1)*REFER_WIDTH)->b);
        }
        /* Lecture de la ligne suivante de l'image droite*/
        for(x = 0 ; x < TRANS_WIDTH; x++)
        {
            getpixel_fast_antialiasing(right, right_norm_data,
                                       x + (int)right_norm_data->left,
                                       y + (int)right_norm_data->top, 0, 0,
                                       &(lines_trans+x+(WINTRANS_Y-1)*TRANS_WIDTH)->r,
                                       &(lines_trans+x+(WINTRANS_Y-1)*TRANS_WIDTH)->g,
                                       &(lines_trans+x+(WINTRANS_Y-1)*TRANS_WIDTH)->b);
        }
        /* On attend d'avoir lu WINTRANS_Y lignes */
        if(y < WINTRANS_Y - 1)
            continue;

	/*Parcours de la ligne*/
	for(x = 0; x < left_norm_data->right-left_norm_data->left; x++)
        {
#if 1
	    /*Correlation*/
            corr = Translation_bande(x,y,
                                     lines_refer + REFER_WIDTH * (CORR_Y/2),
                                     lines_trans + TRANS_WIDTH * (WINTRANS_Y/2),
                                     Map_norm->elem);
	    pt2d.x = x+left_norm_data->left;
	    pt2d.y = y+left_norm_data->top;
	    pt2d = inverse_transform(left_norm_data,pt2d);
	     
	    if(pt2d.x >=0 && pt2d.y >=0) 
	    {
	        (Map->elem+(int)pt2d.y*Map->width+(int)pt2d.x)->depth = 
		 			(Map_norm->elem+REFER_WIDTH+x)->depth;
	        (Map->elem+(int)pt2d.y*Map->width+(int)pt2d.x)->corr = 
		 			(Map_norm->elem+REFER_WIDTH+x)->corr;
		memcpy((Map->elem+(int)pt2d.y*Map->width+(int)pt2d.x)->tabcorr,
		       (Map_norm->elem+REFER_WIDTH+x)->tabcorr,
		       sizeof(int)*TRANS_X);
	    }
#else
	    corr = 0;
       	    pt2d.x = x+left_norm_data->left;
	    pt2d.y = y+left_norm_data->top;
	    pt2d = inverse_transform(left_norm_data,pt2d);
	    if(pt2d.x >=0 && pt2d.y >=0) 
	    {
	        int r,g,b;
//	        (Map->elem+(int)pt2d.y*Map->width+(int)pt2d.x)->depth = (lines_refer+x)->r ;
//	        (Map->elem+(int)pt2d.y*Map->width+(int)pt2d.x)->corr = 0;
		getpixel_fast(left,x,y,&r,&g,&b);
		r = (lines_refer+x)->r;
	        (Map->elem+(int)pt2d.y*Map->width+(int)pt2d.x)->depth =r ;
	        (Map->elem+(int)pt2d.y*Map->width+(int)pt2d.x)->corr = 0;
//	        (Map->elem+(int)(y)*Map->width+(int)x)->depth = r ;
//	        (Map->elem+(int)y*Map->width+(int)x)->corr = 0;
	    }
#endif
        }
	/*Décalage de la seconde à la première ligne*/
	memcpy(Map_norm->elem,Map_norm->elem+REFER_WIDTH,sizeof(pixeldepth_t)*REFER_WIDTH);
    }
    printf("Dimension : %d %d\n",left->width,left->height);
    printf("100 %% effectués\n");
    {
        image_t *lines=malloc(sizeof(image_t));
        image_t *corr_min=malloc(sizeof(image_t));
        image_t *line=malloc(sizeof(image_t));
	int indic;
	int ycorr;
	int pixel;
	//int I=150;
        image_new(lines, left->width,left->height, 3);
        image_new(corr_min, left->width,left->height, 3);
        image_new(line, left->width,TRANS_X*left->height, 3);
	cvFillImage(lines->_private,CV_RGB(255,255,255));
	cvFillImage(corr_min->_private,CV_RGB(255,255,255));
	cvFillImage(line->_private,CV_RGB(0,100,0));
	for(y=0 ; y <left->height ; y++)
	{
  	     for(x=0 ; x <left->width  ; x++)
	     {

	         pixel =(Map->elem+y*Map->width+x)->tabcorr[TRANS_X/2] ;
	         setpixel(lines,x,y,pixel,pixel,pixel);
		   //pixel =(Map->elem+y*Map->width+x)->corr/(9*16*sqrt(3));
	         pixel =fabs((Map->elem+y*Map->width+x)->corr-
		  (Map->elem+y*Map->width+x-1)->corr);
		 setpixel(corr_min,x,y,pixel,pixel,pixel);
	     }
	}
	for(y=0 ; y<left->height; y++)
	{
	    for(x=0 ; x<left->width ; x++)
  	    {
	        for(indic=0 , ycorr=0 ; ycorr<TRANS_X+1 ; ycorr++)
	        {
  		     pixel = (Map->elem+y*Map->width+x)->tabcorr[ycorr];
		     if(pixel == (Map->elem+y*Map->width+x)->corr)
 		         setpixel_fast(line,x,y*TRANS_X+ycorr,255,0,0);
		     else
		     {
 		        //pixel /= 9*16*16*TRANS_X*8*TRANS_X*sqrt(3); 
 		        //pixel /= 9*16*sqrt(3); 
		        pixel /=8;
		        if(pixel > 255)
		          setpixel_fast(line,x,y*TRANS_X+ycorr,0,0,255);
		        else
		        {
		          setpixel_fast(line,x,y*TRANS_X+ycorr,pixel,pixel,pixel);
		        }
		    }
		} 
	    }
	}
	cvSaveImage("line.png",line->_private);
	cvSaveImage("lines.png",lines->_private);
	cvSaveImage("corr.png",corr_min->_private);
    }


    return THRESH;
}
