#ifdef _WIN32
typedef unsigned char uint8_t;
#else
#include <stdint.h>
#endif

#include <cv.h>
#include <stdio.h>

#include "defs.h"
#include "image.h"
#include "main.h"

#include <math.h>

#include <cv.h>
#include <highgui.h>

#include "geometry.h"

#include "correlation.h"

static pt2d_t inverse_transform(projection_t *data,
                                pt2d_t from)
{
    pt2d_t to;
    pt3d_t xx, yy, cm;

    xx = multiply(from.x, data->vx);
    yy = multiply(from.y, data->vy);

    cm = add(data->Os, add(xx, yy));

    to.x = cm.x * data->d / cm.z;
    to.y = cm.y * data->d / cm.z;

    return to;
}
#if 0
static pt2d_t transform(projection_t *data,
                        pt2d_t from)
{
    pt3d_t pt3d;
    pt2d_t to;

    pt3d.x = from.x;
    pt3d.y = from.y;
    pt3d.z = data->d;

    pt3d = multiply(1 / scalar_product(pt3d, data->Os), pt3d);
    pt3d = multiply(scalar_product(data->Os, data->Os), pt3d);

    to.x = scalar_product(pt3d, data->vx);
    to.y = scalar_product(pt3d, data->vy);

    return to;
}
#endif
static int getpixel_fast_aa(image_t *img, projection_t *data,
                            double x, double y, int offsetx,
                            int offsety, int *r, int *g, int *b)
{
    double i0,j0,i1,j1,I_,J_;
    int pR,pB,pG;
    double c1,c2,c3,c4;
    pt2d_t from,to;

    from.x = x;
    from.y = y;

    /*Point normé => point dénormé*/
    to = inverse_transform(data, from);

    /* Pourcentage de couleur */
    j0 = to.x;
    i0 = to.y;
    j1 = to.x + 1.0;
    i1 = to.y + 1.0;

    J_ = (int)to.x;
    I_ = (int)to.y;

    c1 = ((I_+1) - i0) * ((J_+1) - j0);
    c2 = ((I_+1) - i0) * (j1 - (J_+1));
    c3 = (i1 - (I_+1)) * (j1 - (J_+1));
    c4 = (i1 - (I_+1)) * ((J_+1) - j0);
    pR = pB = pG = 0;
    getpixel_fast(img,(int)to.x+offsetx,(int)to.y+offsety,r,g,b);
    pR = c1 * (*r);
    pG = c1 * (*g);
    pB = c1 * (*b);
    getpixel_fast(img,(int)to.x+1+offsetx,(int)to.y+offsety,r,g,b);
    pR += c2 * (*r);
    pG += c2 * (*g);
    pB += c2 * (*b);
    getpixel_fast(img,(int)to.x+1+offsetx,(int)to.y+1+offsety,r,g,b);
    pR += c3*(*r);
    pG += c3*(*g);
    pB += c3*(*b);
    getpixel_fast(img,(int)to.x+offsetx,(int)to.y+1+offsety,r,g,b);
    pR += c4*(*r);
    pG += c4*(*g);
    pB += c4*(*b);

    *r = pR;
    *g = pG;
    *b = pB;

    return 0;
}

/*****************************************************************************
 * Calcul_bande: calcule la distance entre deux matrices de pixels
 *               de taille CORR_X × CORR_Y
 *****************************************************************************/
static int Calcul_bande(window_t *Refer, window_t *Trans,
                        int dx, int dy, int x, int y)
{
    int i, j;
    int corr = 0;
    int delta;
    window_t P1, P2;

    for(j = -CORR_Y/2 ; j < CORR_Y-CORR_Y/2; j++)
    {
        int k = j * REFER_WIDTH + x;
        int l = (j + dy) * TRANS_WIDTH + x + dx;

        for(i = -CORR_X/2 ; i < CORR_X-CORR_X/2; i++)
        {
            if(i+x < 0 || i+x >= REFER_WIDTH-1)
            {
                P1.r = 255;
                P1.g = 0;
                P1.b = 255;
            }
            else
            {
                P1 = *(Refer+k+i);
            }

            if(x+i+dx < 0 || i + x + dx >= TRANS_WIDTH - 1)
            {
                P2.r = 255;
                P2.g = 0;
                P2.b = 255;
            }
            else
            {
                P2 = *(Trans+l+i);
            }

            /* Distance euclidienne */
            delta = (sq(P1.r-P2.r) + sq(P1.g-P2.g) + sq(P1.b-P2.b));
            corr += delta;
        }
    }
    return corr;
}

/*****************************************************************************
 * Translation_bande: calcule la meilleure corrélation d'une matrice de
 *                    taille CORR_X × CORR_Y dans une fenêtre de taille
 *                    TRANS_X × TRANS_Y.
 *****************************************************************************/
static inline void Translation_bande(int x, int y, window_t *Refer,
                                     window_t *Trans, int *tabcorr)
{
    int dx, dy;

    /* Parcours de la translation de gauche a droite */
    for(dy = -TRANS_Y/2; dy < TRANS_Y-TRANS_Y/2; dy++)
        for(dx = -TRANS_X/2; dx < TRANS_X-TRANS_X/2; dx++)
            tabcorr[dx+TRANS_X/2] = Calcul_bande(Refer, Trans, dx, dy, x, y);
}

/*****************************************************************************
 * Correlation_bande2: le gros de la corrélation
 *****************************************************************************/
int Correlation_bande2(image_t *left , projection_t *left_norm_data,
                       image_t *right, projection_t *right_norm_data,
                       image_t *left_depthmap_norm,
                       image_t *right_depthmap_norm, double adjust,map_t *Map)
{
    int x, y, i;
    int progress;
    window_t *lines_refer, *lines_trans;
//    pt2d_t pt2d;
    printf("CORR 2\n");

    /*Détermination de la taille de la bande de référence de 3 lignes*/
//    trans_width = (right_norm_data->right-right_norm_data->left) + 1;
//    refer_width = left_norm_data->right - left_norm_data->left + 1;

    /*Allocation des lignes normalisée de 3 lignes*/
    lines_refer = malloc(sizeof(window_t)*(REFER_WIDTH*CORR_Y+1));
    lines_trans = malloc(sizeof(window_t)*(TRANS_WIDTH*WINTRANS_Y+1));

    /* Initialisation des images résultat en bleu */
    cvFillImage(left_depthmap_norm->_private, CV_RGB(0,0,255));
    cvFillImage(right_depthmap_norm->_private, CV_RGB(0,0,255));

    /* Parcours de l'image normalisée par lignes */
    for(progress = 0, y = 0 ; y < left_norm_data->bottom - left_norm_data->top; y++)
    {
        while( y * 100 / (left_norm_data->bottom - left_norm_data->top) >= progress)
        {
            printf("\r %i %% effectués", progress);
            fflush(stdout);
            progress ++;
        }

        /* Décalage de ligne */
        for(i = 0; i < CORR_Y - 1; i++ )
            memcpy(lines_refer + REFER_WIDTH * i,
                   lines_refer + REFER_WIDTH * (i + 1),
                   sizeof(window_t) * REFER_WIDTH);

        for(i = 0; i < WINTRANS_Y - 1; i++ )
            memcpy(lines_trans + TRANS_WIDTH * i,
                   lines_trans + TRANS_WIDTH * (i + 1),
                   sizeof(window_t) * TRANS_WIDTH);

        /* Lecture de la ligne suivante de l'image gauche */
        for(x = 0 ; x < REFER_WIDTH; x++)
        {
            getpixel_fast_aa(left, left_norm_data,
                             (int)left_norm_data->left + x,
                             (int)left_norm_data->top + y + adjust,
                             0, 0,
                             &(lines_refer+x+(CORR_Y-1)*REFER_WIDTH)->r,
                             &(lines_refer+x+(CORR_Y-1)*REFER_WIDTH)->g,
                             &(lines_refer+x+(CORR_Y-1)*REFER_WIDTH)->b);
        }

        /* Lecture de la ligne suivante de l'image droite*/
        for(x = 0 ; x < TRANS_WIDTH; x++)
        {
            getpixel_fast_aa(right, right_norm_data,
                             (int)right_norm_data->left + x,
                             (int)right_norm_data->top + y, 0, 0,
                             &(lines_trans+x+(WINTRANS_Y-1)*TRANS_WIDTH)->r,
                             &(lines_trans+x+(WINTRANS_Y-1)*TRANS_WIDTH)->g,
                             &(lines_trans+x+(WINTRANS_Y-1)*TRANS_WIDTH)->b);
        }

        /* On attend d'avoir lu WINTRANS_Y lignes */
        if(y < WINTRANS_Y - 1)
            continue;

        /* Parcours de la ligne */
        for(x = 0; x < left_norm_data->right - left_norm_data->left; x++)
        {
            /* Correlation */
            Translation_bande(x, y, lines_refer + REFER_WIDTH * (CORR_Y/2),
                                    lines_trans + TRANS_WIDTH * (WINTRANS_Y/2),
                                    ((Map->elem+y*Map->width+x)->tabcorr));
        }
    }
    printf("\r 100%% effectués\n");

    printf("\r100%% effectués\n");

    printf("Calcul de la carte de profondeur\n");

    /* Calcul de toutes les lignes de profondeur */
    for(progress = 0, y = 0 ; y < Map->height ; y++)
    {
        while( y * 100 / Map->height >= progress)
        {
            printf("\r %i %% effectués", progress);
            fflush(stdout);
            progress ++;
        }

#if 0
        /* Truc pas beau */
        for(x = 2; x < Map->width; x++)
        {
             pixeldepth_t *elem = Map->elem + y * Map->width + x;
             int dy;

             for(dy = 0 ; dy < TRANS_X ; dy++)
             {
                 (elem - 2)->tabcorr[dy] += elem->tabcorr[dy] / 8;
                 (elem - 2)->tabcorr[dy] += (elem -1)->tabcorr[dy] / 4;
             }
        }
#endif

        /* Recherche des minima pour chaque X */
        for(x = 0; x < Map->width; x++)
        {
             pixeldepth_t *elem = Map->elem + y * Map->width + x;
             int dy;

             elem->corr = 99999999;
             elem->corr1 = 99999999;
             elem->corr2 = 99999999;
             elem->invalid = 0;
             for(dy = 0 ; dy < TRANS_X ; dy++)
             {
                 if(elem->tabcorr[dy] < elem->corr)
                 {
                     elem->corr2 = elem->corr1;
                     elem->depth2 = elem->depth1;
                     elem->corr1 = elem->corr;
                     elem->depth1 = elem->depth;
                     elem->corr = elem->tabcorr[dy];
                     elem->depth = dy - TRANS_X / 2;
                 }
                 else if(elem->tabcorr[dy] < elem->corr1)
                 {
                     elem->corr2 = elem->corr1;
                     elem->depth2 = elem->depth1;
                     elem->corr1 = elem->tabcorr[dy];
                     elem->depth1 = dy - TRANS_X / 2;
                 }
                 else if(elem->tabcorr[dy] < elem->corr2)
                 {
                     elem->corr2 = elem->tabcorr[dy];
                     elem->depth2 = dy - TRANS_X / 2;
                 }
             }
        }

#if 1
        /* Amélioration de la courbe de minima */
        for(x = 0; x < Map->width - 4; x++)
        {
            pixeldepth_t *elem = Map->elem + y * Map->width + x;

            if(elem[0].invalid || elem[1].invalid)
                continue;

            /* Lissage tout bête */
            if(elem[0].depth == elem[1].depth
                && elem[0].depth != elem[2].depth
                && elem[0].depth == elem[3].depth)
            {
                elem[2].depth = elem[0].depth;
                //elem[2].highlight = 1;
            }
            else if(elem[0].depth != elem[1].depth
                     && elem[0].depth == elem[2].depth
                     && elem[0].depth == elem[3].depth)
            {
                elem[1].depth = elem[0].depth;
                //elem[1].highlight = 1;
            }
            /* Lissage un peu plus bourrin */
            else if(elem[0].depth == elem[1].depth
                     && elem[0].depth != elem[2].depth
                     && elem[0].depth != elem[3].depth
                     && elem[0].depth == elem[4].depth)
            {
                elem[2].depth = elem[0].depth;
                //elem[2].highlight = 1;
                elem[3].depth = elem[0].depth;
                //elem[3].highlight = 1;
            }
            else if(elem[0].depth != elem[1].depth
                     && elem[0].depth != elem[2].depth
                     && elem[0].depth == elem[3].depth
                     && elem[0].depth == elem[4].depth)
            {
                elem[1].depth = elem[0].depth;
                //elem[1].highlight = 1;
                elem[2].depth = elem[0].depth;
                //elem[2].highlight = 1;
            }
#if 1
            /* Élagage des paires de points aberrants */
            else if(abs(elem[0].depth - elem[2].depth) > 2
                     && abs(elem[0].depth - elem[3].depth) > 2
                     && abs(elem[1].depth - elem[2].depth) > 2
                     && abs(elem[1].depth - elem[3].depth) > 2
                     && abs(elem[4].depth - elem[2].depth) > 2
                     && abs(elem[4].depth - elem[3].depth) > 2)
            {
                int mean = (elem[0].depth + elem[1].depth
                                          + elem[4].depth) / 3;
                int mini = min(elem[0].depth,
                               min(elem[1].depth, elem[4].depth));
                if(abs(mini - mean) < 2)
                {
                    elem[2].depth = mean;
//                    elem[2].highlight = 1;
                    elem[3].depth = mean;
//                    elem[3].highlight = 1;
                }
                else
                {
                    elem[2].invalid = 1;
                    elem[3].invalid = 1;
                }
            }
            /* Élagage des points aberrants */
            else if(abs(elem[0].depth - elem[1].depth) > 2
                     && abs(elem[2].depth - elem[1].depth) > 2
                     /*&& abs(elem[3].depth - elem[1].depth) > 2*/)
            {
                int mean = (elem[0].depth + elem[2].depth
                                          ) / 2;
                                          //+ elem[3].depth) / 3;
                int mini = min(elem[0].depth,
                               elem[2].depth);
                               //min(elem[2].depth, elem[3].depth));
                if(abs(mini - mean) < 2)
                {
                    elem[1].depth = mean;
                    //elem[1].highlight = 1;
                }
                else
                    elem[1].invalid = 1;
            }
#endif
        }
#endif

#if 0
        /* Do the diagonals */
        for(x = -DELTA_X; x < Map->width; x++)
        {
            int dx;
            int old_dx, old_corr;

            /* Find the first valid point of the diagonal */
            for(dx = (x < 0) ? -x : 0;
                dx < DELTA_X + 1 && x + dx < Map->width;
                dx++)
            {
//                    if(x == 49 || x == 52)
//                        (Map->elem + y * Map->width + x + dx)->tabcorr[TRANS_X - 1 - dx] = 100000;
                if((Map->elem + y * Map->width + x + dx)->depth + TRANS_X / 2
                     == TRANS_X - 1 - dx)
                    break;
            }

            /* If we didn't find any, just bail off */
            if(dx == DELTA_X + 1 || x + dx >= Map->width)
                continue;

//if(x != -21 /*&& x != -20*/)
  //  continue;

//if(x&1)
//(Map->elem + y * Map->width + x + dx)->invalid = 1;
//else
//(Map->elem + y * Map->width + x + dx)->highlight = 1;
//continue;
            old_dx = dx;
            old_corr = (Map->elem + y * Map->width + x + dx)->corr;

#if 1
            for(dx++; dx < TRANS_X && x + dx < Map->width; dx++)
            {
                if((Map->elem + y * Map->width + x + dx)->depth + TRANS_X / 2
                     == TRANS_X - 1 - dx)
                {
                    if(old_corr >= (Map->elem + y * Map->width + x + dx)->corr)
                    {
                        (Map->elem + y * Map->width + x + dx)->invalid = 1;
                    }
                    else
                    {
                        (Map->elem + y * Map->width + x + old_dx)->invalid = 1;
                        old_corr = (Map->elem + y * Map->width + x + dx)->corr;
                        old_dx = dx;
                    }
                }
            }
#endif
        }
#endif
    }
    printf("\r 100%% effectués\n");
    printf("\r100%% effectués\n");

#if 0
    {
        int ycorr;
        int pixel;
        printf("Sauvegarde de l'image des corrélations...");
	image_t *line=malloc(sizeof(image_t));
        image_new(line,left_depthmap_norm->width,left_depthmap_norm->height*TRANS_X,3);
        cvFillImage(line->_private,CV_RGB(0,0,255));
        for(y=0 ; y<left_depthmap_norm->height ; y++)
        {
            for(x=0 ; x<left_depthmap_norm->width ; x++)
            {
                for(ycorr = 0; ycorr < TRANS_X; ycorr++)
                {
                   int xpix = x, ypix = y * TRANS_X + (TRANS_X - 1 - ycorr);
                   pixel = (Map->elem+y*Map->width+x)->tabcorr[ycorr];
                   if(ycorr == (Map->elem+y*Map->width+x)->depth + TRANS_X / 2)
                   {
                       if((Map->elem+y*Map->width+x)->highlight)
                           setpixel_fast(line, xpix, ypix, 255, 255, 0);
                       else if((Map->elem+y*Map->width+x)->invalid)
                           setpixel_fast(line, xpix, ypix, 0, 255, 0);
                       else
                           setpixel_fast(line, xpix, ypix, 255, 0, 0);
                   }
                   else
                   {
                       //pixel *= 2;
                       pixel /= 128;
                       if(pixel > 255)
                           //setpixel_fast(line, xpix, ypix, 255, 255, 255);
                           setpixel_fast(line, xpix, ypix, 0, 0, 255);
                       else
                           setpixel_fast(line, xpix, ypix, pixel,pixel,pixel);
                   }
                }
            }
        }
        cvSaveImage("line.bmp",line->_private);
	image_free(line);
	free(line);
	printf("...ok\n");
    }
#endif

    free(lines_refer);
    free(lines_trans);
    return THRESH * 20;
}


