/*
 *  stereo -- Stereo matching software
 *  $Id: image.c 306 2004-09-09 13:24:53Z sam $
 *  Copyright (c) 2004 Sam Hocevar <sam@zoy.org>
 *                     Éric Beets <ericbeets@free.fr>
 *  All Rights Reserved
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA
 */

#include <stdio.h>
#include <stdint.h>

#include <cv.h>
#include <highgui.h>

#include "defs.h"
#include "image.h"
/* Lecture des fichiers TGA*/
static int freadint(FILE *fp)
{
	unsigned char buffer[4];
	fread(buffer, 4, 1, fp);
	return ((int)buffer[3] << 24) | ((int)buffer[2] << 16) | ((int)buffer[1] << 8) | ((int)buffer[0]);
}
                                                                                
static int freadshort(FILE *fp)
{
	unsigned char buffer[2];
	fread(buffer, 2, 1, fp);
	return ((int)buffer[1] << 8) | ((int)buffer[0]);
}
                                                                                
static int freadchar(FILE *fp)
{
	unsigned char buffer;
	fread(&buffer, 1, 1, fp);
	return (int)buffer;
}
/*
 * Lecture des fichiers TGA 
 */
char * read_tga(char *name,int *pw,int *ph,int *pdepth)
{
	char *pixels = NULL;
	FILE *fp;
	static int tmp;
	int bpp, w, h, step, depth;
	int rmask,gmask,bmask,amask,i;
	fp = fopen(name, "rb");
	if(!fp)
	{
		printf("Erreur chargement TGA...\n");
		return NULL;
	}
	                                                                                
	tmp = freadshort(fp);
	fseek(fp, 0, SEEK_SET);
	                                                                                
	if(tmp == 0x4d42)
	{
		/* Read a BMP file */
	}
	else
	{
		/* Read a TGA file */
	                                                                                
		tmp = freadchar(fp);		/* ID length */
		if(freadchar(fp) != 0x00)	/* Colormap type */
		{
			fclose(fp);
			return NULL;
		}
		if(freadchar(fp) != 0x02)	/* Image type */
		{
			fclose(fp);
			return NULL;
		}
	                                                                                
		/* Skip Colormap specification */
		fseek(fp, 5, SEEK_CUR);
	                                                                                
		/* Read Image specification */
		w = freadshort(fp); /* X origin */
		h = freadshort(fp); /* Y origin */
		w = freadshort(fp);
		h = freadshort(fp);
		bpp = freadchar(fp);
		freadchar(fp); /* image descriptor */
	                                                                                
		/* Skip image ID */
		fseek(fp, tmp, SEEK_CUR);
	                                                                                
		/* Fill in more stuff */
		depth = (bpp + 7) / 8;
		step = (w * depth + 3) & ~3;

		switch(depth)
		{
			case 3:
				rmask = 0xff0000;
				gmask = 0x00ff00;
				bmask = 0x0000ff;
				amask = 0x000000;
				break;
			case 2: /* XXX: those are the 16 bits values */
				rmask = 0x7c00;
				gmask = 0x03e0;
				bmask = 0x001f;
				amask = 0x0000;
				break;
			case 1:
			default:
				bpp = 8;
				rmask = gmask = bmask = amask = 0;
				break;
		}
	
		/* Allocate the pixel buffer */
		pixels = malloc(w * h * depth);
		if(!pixels)
		{
			fclose(fp);
			return NULL;
		}
		memset(pixels, 0, w * h * depth);
	
		for(i = h; i--; /* */)
			fread(pixels + step * i, w * depth, 1, fp);

		*pdepth = depth;
		*pw = w;
		*ph = h;
	}
	
	fclose(fp);
	return pixels;
}


int image_new_16(image_t *image,unsigned int width,unsigned int height,unsigned int depth)
{
    IplImage *tmp;

    image->width = width;
    image->height = height;
    image->pitch = (width * depth + 15) & ~15;
    image->depth = depth;

# if 0
    image->pixels = malloc(height * image->pitch);
    if(!image->pixels) return -1;

    image->_type = IMAGE_NORMAL;
#else
    tmp = cvCreateImage(cvSize(width, height), IPL_DEPTH_16S, depth);
    cvFillImage(tmp,CV_RGB(255,255,255));
    image->width = tmp->width;
    image->height = tmp->height;
    image->pitch = tmp->widthStep;
    image->depth = tmp->nChannels;
    image->pixels = tmp->imageData;

    image->_private = (void *)tmp;
    image->_type = IMAGE_OPENCV;
#endif

    return 0;
}

int image_new(image_t *image, unsigned int width, unsigned int height, unsigned int depth)
{
    IplImage *tmp;

    image->width = width;
    image->height = height;
    image->pitch = (width * depth + 15) & ~15;
    image->depth = depth;

# if 0
    image->pixels = malloc(height * image->pitch);
    if(!image->pixels) return -1;

    image->_type = IMAGE_NORMAL;
#else
    tmp = cvCreateImage(cvSize(width, height), IPL_DEPTH_8U, depth);

    image->width = tmp->width;
    image->height = tmp->height;
    image->pitch = tmp->widthStep;
    image->depth = tmp->nChannels;
    image->pixels = tmp->imageData;

    image->_private = (void *)tmp;
    image->_type = IMAGE_OPENCV;
#endif

    return 0;
}
IplImage *ChartoIpl(char *pixels,int w,int h,int depth)
{
	IplImage *tmp;
	int step = (w*3 + 3) & ~3;
	tmp = cvCreateImageHeader (cvSize(w,h), IPL_DEPTH_8U, 3);
//	step = (w*depth);
	cvInitImageHeader(tmp, cvSize(w,h), IPL_DEPTH_8U, 3 /* chan */,
			IPL_ORIGIN_BL, 4 /* align */);
	cvSetImageData(tmp, pixels, w * 3);
	return tmp;
}

int image_load(image_t *image, char *filename)
{
    IplImage *tmp;
    int type;

    if(strstr(filename,".tga")!=NULL)
    {
        char *pixels;
	int w,h,depth;
        pixels = read_tga(filename,&w,&h,&depth);
	if(pixels)
	{
	    tmp = ChartoIpl(pixels,w,h,depth);
	  //  free(pixels);
            type = IMAGE_OPENCV_HEADER;
	}
	else
	{
	    printf("erreur de chargement\n");
	    return -1;
	}
    }
    else
    {
        tmp = (void *)cvLoadImage(filename, -1);
        if(!tmp) return -1;

        if(tmp->nChannels != 3)
        {
            IplImage *color;
       
            color = cvCreateImage(cvSize(tmp->width, tmp->height), IPL_DEPTH_8U, 3);
            if(!color)
            {
                cvReleaseImage(&tmp);
                return -1;
            }
            cvCvtColor(tmp, color, CV_GRAY2BGR);
            cvReleaseImage(&tmp);
            tmp = color;
        }

        type = IMAGE_OPENCV;
    }
    image->width = tmp->width;
    image->height = tmp->height;
    image->pitch = tmp->widthStep;
    image->depth = tmp->nChannels;
    image->pixels = tmp->imageData;

    image->_private = (void *)tmp;
    image->_type = type;

    return 0;
}

int image_display(image_t *image)
{
    char name[BUFSIZ];
    IplImage *tmp;

    if(image->_type != IMAGE_OPENCV)
        return -1;

    tmp = (IplImage *)image->_private;
    sprintf(name, "Image %p (%i x %i)", tmp, image->width, image->height);
    cvNamedWindow(name, 0);
    cvShowImage(name, tmp);
    cvResizeWindow(name,640,480);
    return 0;
}

int image_resize(image_t *img,float factor)
{
   image_t *tmp;
   tmp = malloc(sizeof(image_t));
   image_new(tmp,img->width*factor,img->height*factor,3);
   
   cvResize(img->_private,tmp->_private,CV_INTER_NN);
   image_free(img);
   image_copy(img,tmp);
   return 0;
}

int image_save(char *file,image_t *image)
{
    cvSaveImage(file,image->_private);
    return 0;
}
int image_free(image_t *image)
{
    if(!image)
     return 0;
    switch(image->_type)
    {
        case IMAGE_NORMAL:
            free(image->pixels);
            break;
        case IMAGE_OPENCV:
        {
            IplImage *tmp;
            tmp = (IplImage *)image->_private;
            cvReleaseImage(&tmp);
            break;
        }
        case IMAGE_OPENCV_HEADER:
        {
            IplImage *tmp;
            tmp = (IplImage *)image->_private;
            cvReleaseImageHeader(&tmp);
            free(image->pixels);
            break;
        }
        default:
            break;
    }

    image->_type = 0;

    return 0;
}

int getpixel(image_t *image, int x, int y, int *r, int *g, int *b)
{
    if(x < 0 || y < 0 || x >= image->width || y >= image->height)
    {
        *r = 255;
        *g = 0;
        *b = 255;
        return -1;
    }

    *b = image->pixels[y * image->pitch + x * image->depth];
    *g = image->pixels[y * image->pitch + x * image->depth + 1];
    *r = image->pixels[y * image->pitch + x * image->depth + 2];

    return 0;
}

int setpixel(image_t *image, int x, int y, int r, int g, int b)
{
    if(x < 0 || y < 0 || x >= image->width || y >= image->height)
        return -1;

    image->pixels[y * image->pitch + x * image->depth] = b;
    image->pixels[y * image->pitch + x * image->depth + 1] = g;
    image->pixels[y * image->pitch + x * image->depth + 2] = r;

    return 0;
}

int convert_BGR2GRAY(image_t *src,image_t *dst)
{
	image_new(dst,src->width,src->height,1);
	acp(src,dst);
	return 0;
}
int convert_BGR2RGB(image_t *src,image_t *dst)
{
	int x,y;
	int R,G,B;
	for(y=0 ; y<src->height; y++)
	{
		for(x=0 ; x<src->width; x++)
		{
			getpixel(src,x,y,&R,&G,&B);
			setpixel(dst,x,y,B,G,R);
		}
	}
	return 0;
}

int image_copy(image_t* dst,image_t *src)
{
   dst->width  = src->width;
   dst->height = src->height;
   dst->pitch  = src->pitch;
   dst->depth  = src->depth;
   dst->_type= src->_type;

   image_new(dst,src->width,src->height,3);
   cvCopyImage(src->_private,dst->_private);
   return 0;
}

