/*
 *  stereo -- Stereo matching software
 *  $Id: matching.c 303 2004-09-09 12:01:14Z eric $
 *  Copyright (c) 2004 Sam Hocevar <sam@zoy.org>
 *                     Éric Beets <ericbeets@free.fr>
 *  All Rights Reserved
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA
 */

#include <stdint.h>
#include <stdio.h>

#include "defs.h"
#include "main.h"
#include "image.h"

#include "correlation.h"
#include <math.h>
typedef enum _matching_mode matching_mode;

enum _matching_mode
{
	delaunay,
	corr_base,
	corr_normal,
	corr_vole,
	corr_bande,
	corr_bande2
};

matching_mode Matching_mode = corr_bande2;
//matching_mode Matching_mode = delaunay;
/* Perform stereo matching of an image pair */
int matching(pair_t *pair)
{
    int x,y;
    double min, max;
    map_t *Map;
    int seuil=0;

#if 1
    /*Allocation de la structure 'Map' de sortie normalisée*/
    Map = malloc(sizeof(map_t));
//    Map->width = pair->image_norm[LEFT].width+1 ;
//    Map->height = pair->image_norm[LEFT].height+1;
    Map->width = pair->projection[LEFT].right - pair->projection[LEFT].left+1;
    Map->height= pair->projection[LEFT].bottom - pair->projection[LEFT].top+1;
    Map->elem = malloc(sizeof(pixeldepth_t) * Map->height * Map->width);
    for(y=0 ; y<Map->height ; y++)
    {
	for(x=0 ; x<Map->width ; x++)
	{
	  //memset((Map->elem+x+Map->width*y)->tabcorr, 0, sizeof(int)*TRANS_X);
	    memset((Map->elem+x+Map->width*y), 0, sizeof(pixeldepth_t));
	}
    }
    /*Allocation de la carte de profondeur normlisé*/
    image_new(&pair->depthmap_norm[LEFT], Map->width,
              Map->height, 3);
    image_new(&pair->depthmap_norm[RIGHT], 
              pair->projection[RIGHT].right -pair->projection[RIGHT].left,
              pair->projection[RIGHT].bottom-pair->projection[RIGHT].top, 3);
#else
    /*Allocation de la structure 'Map' de sortie dénormalisée*/
    Map = malloc(sizeof(map_t));
    Map->width  = pair->image[LEFT].width;
    Map->height = pair->image[LEFT].height;
    Map->elem = calloc(Map->height*Map->width,sizeof(pixeldepth_t));
    for(y=0 ; y<Map->height ; y++)
    {
	for(x=0 ; x<Map->width ; x++)
	{
	    memset((Map->elem+x+Map->width*y)->tabcorr, 0, sizeof(int)*TRANS_X);
	}
    }
    /*Allocation de la carte de profondeur dénormlisé*/
    image_new(&pair->depthmap_norm[LEFT], pair->image[LEFT].width,
              pair->image[LEFT].height, 3);
    image_new(&pair->depthmap_norm[RIGHT], pair->image[RIGHT].width,
              pair->image[RIGHT].height, 3);
#endif
    printf("Dimension normalisée   : %d %d\n",Map->width,
      					      Map->height);
    printf("Dimension dénormalisée : %d %d\n",pair->image[LEFT].width,
      			  		      pair->image[LEFT].height);
    
   
    /*Allocation de la carte de profondeur dénormalisée*/
    image_new(&pair->depthmap[LEFT], pair->image[LEFT].width,
              pair->image[LEFT].height, 3);
    image_new(&pair->depthmap[RIGHT], pair->image[RIGHT].width,
              pair->image[RIGHT].height, 3);

//	return 0;
    
	printf("Construction carte 3D\n");
	printf("Résolution  : %d\n",RESOLUTION);
	printf("Zone refer  : %d %d\n",CORR_X,CORR_Y);
	printf("Zone recher.: %d %d\n",TRANS_X,TRANS_Y);
	printf("Continuité  : %d %d (vert/horiz)\n",X_WEIGHT,Y_WEIGHT);
	switch(Matching_mode)
	{
		case delaunay:
			Carte_3D(&pair->projection[LEFT],
				 &pair->projection[RIGHT],
				 &pair->depthmap_norm[LEFT],
				 &pair->depthmap_norm[RIGHT]);
			break;
		case corr_base:
			Correlation_base(&pair->image_norm[LEFT],
					 &pair->image_norm[RIGHT],
					 &pair->depthmap[LEFT],
					 &pair->depthmap[RIGHT],
					 pair->adjust);
			break;
		case corr_normal:
			Correlation_normal(&pair->image_norm[LEFT],
					   &pair->image_norm[RIGHT],
					   &pair->depthmap_norm[LEFT],
					   &pair->depthmap_norm[RIGHT],
					   pair->adjust);
			break;
		case corr_vole:
			Correlation_vole(&pair->image[LEFT], 
			  	  	 &pair->projection[LEFT],
					 &pair->image[RIGHT],
					 &pair->projection[RIGHT],
					 &pair->depthmap_norm[LEFT],
					 &pair->depthmap_norm[RIGHT],
					 pair->adjust);
			break;
		case corr_bande:
			seuil = Correlation_bande(&pair->image[LEFT], 
			  			  &pair->projection[LEFT],
						  &pair->image[RIGHT],
						  &pair->projection[RIGHT],
						  &pair->depthmap_norm[LEFT],
						  &pair->depthmap_norm[RIGHT],
						  pair->adjust,Map);
			break;
		case corr_bande2:
			seuil = Correlation_bande2(&pair->image[LEFT], 
			  			  &pair->projection[LEFT],
						  &pair->image[RIGHT],
						  &pair->projection[RIGHT],
						  &pair->depthmap_norm[LEFT],
						  &pair->depthmap_norm[RIGHT],
						  pair->adjust,Map);
			break;
	}
#if 0
	{
int p,q,r,g,b,pixel;
    /*Normalisation de l'affichage*/
    for(min = 999999, max = -999999, p = 0 ; p <pair->depthmap_norm[LEFT].width ; p++)
    {
        for(q = 0; q < pair->depthmap_norm[LEFT].height ; q++)
        {
            getpixel(&pair->depthmap_norm[LEFT], p, q, &r, &g, &b);
            if(r == g)
            {
                if(r < min)
                    min = r;
                if(/*r < 20 &&*/ r > max)
                    max = r;
            }
        }
    }
    printf("min/max: %g %g\n",min,max);
    for(p = 0; p < pair->depthmap_norm[LEFT].width ; p++)
    {
        for(q = 0; q < pair->depthmap_norm[LEFT].height; q++)
        {
            getpixel(&pair->depthmap_norm[LEFT], p, q, &r, &g, &b);
            if(r == g)
            {
                pixel = ((float)(max-r)/(float)(max-min))*255.;
                setpixel(&pair->depthmap_norm[LEFT], p, q, pixel, pixel, pixel);
            }
        }
    }
	}

	
return 0;
#endif
#if 1
    /* Normalisation de l'affichage */ 
    printf("parcours de normalisation : %d %d\n",Map->height,Map->width);
    min = 999999; max = -999999;
    for(y=0 ; y<Map->height; y++)
    {
        for(x=0 ; x < Map->width; x++)
        {
#if 0
       //     if( (Map->elem+y*Map->width+x)->corr/(9*16*sqrt(3)) <=  5 &&
//	        fabs((Map->elem+y*Map->width+x)->depth - (Map->elem+y*Map->width+x-1)->depth) < 127 )
	    if( fabs((Map->elem+y*Map->width+x)->depth-(Map->elem+y*Map->width+x-1)->depth) >= 127  )

#else
            if((Map->elem+y*Map->width+x)->corr < seuil)
#endif
            {
                if((Map->elem+y*Map->width+x)->depth < min)
                    min = (Map->elem+y*Map->width+x)->depth;

                if((Map->elem+y*Map->width+x)->depth > max)
                    max = (Map->elem+y*Map->width+x)->depth;
            }
        }
    }
    printf("min/max: %g %g\n",min,max);

    /* Conversion carte de profondeur en image */
    
    for(y=0 ; y<Map->height; y++)
    {
        for(x=0; x<Map->width; x++)
        {
#if 0
//            if( (Map->elem+y*Map->width+x)->corr/(9*16*sqrt(3)) >  5 &&
//	        fabs((Map->elem+y*Map->width+x)->depth - (Map->elem+y*Map->width+x-1)->depth) > 127 )
	    if( fabs((Map->elem+y*Map->width+x)->depth-(Map->elem+y*Map->width+x-1)->depth) < 127  )
	    {
                setpixel_fast(&pair->depthmap_norm[LEFT],x,y,0,0,255);

	    }
#else	 
            if((Map->elem+y*Map->width+x)->invalid)
                setpixel_fast(&pair->depthmap_norm[LEFT],x,y,0,255,0);
            else if((Map->elem+y*Map->width+x)->highlight)
                setpixel_fast(&pair->depthmap_norm[LEFT],x,y,255,255,0);
            else if((Map->elem+y*Map->width+x)->corr > seuil)
                setpixel_fast(&pair->depthmap_norm[LEFT],x,y,255,0,0);
#endif
            else
            {
                int pixel = ((float)(max - (Map->elem+y*Map->width+x)->depth)
                              / (float)(max - min)) * 255.;
                setpixel(&pair->depthmap_norm[LEFT], x, y, 
		         pixel, 
			 pixel, 
			 pixel);
            }
        }
    }
#endif
     /* Sauvegarde des images resultats */
    image_save("sortie_bande.bmp", &pair->depthmap_norm[LEFT]);
    pair->correlation[LEFT].Map = Map;
    return 0;
}

